@@ -14,15 +14,16 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
+#include <config.h>
 #include <stdio.h>
 #include <string.h>
 
 #include "gconf-client.h"
-#include <gconf/gconf-internals.h>
+#include "gconf/gconf-internals.h"
 
 #include "gconfmarshal.h"
 #include "gconfmarshal.c"
@@ -34,7 +35,6 @@ trace (const char *format, ...)
 {
   va_list args;
   gchar *str;
-  FILE *out;
 
   if (!do_trace)
     return;
@@ -45,10 +45,7 @@ trace (const char *format, ...)
   str = g_strdup_vprintf (format, args);
   va_end (args);
 
-  out = stderr;
-  
-  fputs ("gconf trace: ", out);
-  fputs (str, out);
+  g_message ("%s", str);
   
   g_free (str);
 }
@@ -233,6 +230,8 @@ gconf_client_init (GConfClient *client)
   client->error_mode = GCONF_CLIENT_HANDLE_UNRETURNED;
   client->dir_hash = g_hash_table_new (g_str_hash, g_str_equal);
   client->cache_hash = g_hash_table_new (g_str_hash, g_str_equal);
+  client->cache_dirs = g_hash_table_new_full (g_str_hash, g_str_equal,
+					      g_free, NULL);
   /* We create the listeners only if they're actually used */
   client->listeners = NULL;
   client->notify_list = NULL;
@@ -249,7 +248,7 @@ destroy_dir_foreach_remove(gpointer key, gpointer value, gpointer user_data)
   
   if (d->notify_id != 0)
     {
-      trace ("Removing notify ID %u from engine\n", d->notify_id);
+      trace ("REMOTED: Removing notify ID %u from engine", d->notify_id);
       PUSH_USE_ENGINE (client);
 	  gconf_engine_notify_remove (client->engine, d->notify_id);
       POP_USE_ENGINE (client);
@@ -310,6 +309,9 @@ gconf_client_finalize (GObject* object)
   g_hash_table_destroy (client->cache_hash);
   client->cache_hash = NULL;
 
+  g_hash_table_destroy (client->cache_dirs);
+  client->cache_dirs = NULL;
+
   unregister_client (client);
 
   set_engine (client, NULL);
@@ -325,7 +327,7 @@ gconf_client_finalize (GObject* object)
 static void
 gconf_client_real_unreturned_error (GConfClient* client, GError* error)
 {
-  trace ("Unreturned error '%s'\n", error->message);
+  trace ("Unreturned error '%s'", error->message);
   if (client->error_mode == GCONF_CLIENT_HANDLE_UNRETURNED)
     {
       if (global_error_handler != NULL)
@@ -350,7 +352,7 @@ gconf_client_real_unreturned_error (GConfClient* client, GError* error)
 static void
 gconf_client_real_error            (GConfClient* client, GError* error)
 {
-  trace ("Error '%s'\n", error->message);
+  trace ("Error '%s'", error->message);
   if (client->error_mode == GCONF_CLIENT_HANDLE_ALL)
     {
       if (global_error_handler != NULL)
@@ -400,7 +402,7 @@ notify_from_server_callback (GConfEngine* conf, guint cnxn_id,
   g_return_if_fail (GCONF_IS_CLIENT(client));
   g_return_if_fail (client->engine == conf);
 
-  trace ("Received notify of change to '%s' from server\n",
+  trace ("Received notify of change to '%s' from server",
          entry->key);
   
   /* First do the caching, so that state is sane for the
@@ -546,7 +548,7 @@ gconf_client_add_dir     (GConfClient* client,
 
   g_return_if_fail (gconf_valid_key (dirname, NULL));
 
-  trace ("Adding directory '%s'\n", dirname);
+  trace ("Adding directory '%s'", dirname);
   
   d = g_hash_table_lookup (client->dir_hash, dirname);
 
@@ -561,7 +563,7 @@ gconf_client_add_dir     (GConfClient* client,
        */
       if (overlap_dir == NULL)
         {
-          trace ("Adding notify to engine at '%s'\n",
+          trace ("REMOTE: Adding notify to engine at '%s'",
                  dirname);
           PUSH_USE_ENGINE (client);
           notify_id = gconf_engine_notify_add (client->engine,
@@ -626,6 +628,8 @@ foreach_add_notifies(gpointer key, gpointer value, gpointer user_data)
        * already add a notify */
       if (overlap_dir == NULL)
         {
+          trace ("REMOTE: Adding notify to engine at '%s'",
+                 dir->name);
           PUSH_USE_ENGINE (client);
           dir->notify_id = gconf_engine_notify_add(client->engine,
                                                    dir->name,
@@ -656,6 +660,19 @@ clear_dir_cache_foreach (char* key, GConfEntry* entry, char *dir)
     return FALSE;
 }
 
+static gboolean
+clear_cache_dirs_foreach (char *key, gpointer value, char *dir)
+{
+  if (strcmp (dir, key) == 0 ||
+      gconf_key_is_below (dir, key))
+    {
+      trace ("'%s' no longer fully cached", dir);
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
 static void
 gconf_client_real_remove_dir    (GConfClient* client,
                                  Dir* d,
@@ -672,7 +689,7 @@ gconf_client_real_remove_dir    (GConfClient* client,
   
   if (d->notify_id != 0)
     {
-      trace ("Removing notify from engine at '%s'\n", d->name);
+      trace ("REMOTE: Removing notify from engine at '%s'", d->name);
       PUSH_USE_ENGINE (client);
       gconf_engine_notify_remove (client->engine, d->notify_id);
       POP_USE_ENGINE (client);
@@ -682,7 +699,9 @@ gconf_client_real_remove_dir    (GConfClient* client,
   g_hash_table_foreach_remove (client->cache_hash,
                                (GHRFunc)clear_dir_cache_foreach,
                                d->name);
-
+  g_hash_table_foreach_remove (client->cache_dirs,
+                               (GHRFunc)clear_cache_dirs_foreach,
+                               d->name);
   dir_destroy(d);
 
   ad.client = client;
@@ -700,7 +719,7 @@ gconf_client_remove_dir  (GConfClient* client,
 {
   Dir* found = NULL;
 
-  trace ("Removing directory '%s'\n", dirname);
+  trace ("Removing directory '%s'", dirname);
   
   found = g_hash_table_lookup (client->dir_hash,
                                dirname);
@@ -712,7 +731,7 @@ gconf_client_remove_dir  (GConfClient* client,
       found->add_count -= 1;
 
       if (found->add_count == 0) 
-        gconf_client_real_remove_dir(client, found, err);
+        gconf_client_real_remove_dir (client, found, err);
     }
 #ifndef G_DISABLE_CHECKS
   else
@@ -804,12 +823,12 @@ gconf_client_clear_cache(GConfClient* client)
   g_return_if_fail(client != NULL);
   g_return_if_fail(GCONF_IS_CLIENT(client));
 
-  trace ("Clearing cache\n");
+  trace ("Clearing cache");
   
   g_hash_table_foreach_remove (client->cache_hash, (GHRFunc)clear_cache_foreach,
                                client);
 
-  g_assert (g_hash_table_size(client->cache_hash) == 0);
+  g_hash_table_remove_all (client->cache_dirs);
 }
 
 static void
@@ -828,6 +847,7 @@ recurse_subdir_list(GConfClient* client, GSList* subdirs)
       
       cache_pairs_in_dir(client, s);
 
+      trace ("REMOTE: All dirs at '%s'", s);
       PUSH_USE_ENGINE (client);
       recurse_subdir_list(client,
                           gconf_engine_all_dirs (client->engine, s, NULL));
@@ -901,7 +921,7 @@ cache_pairs_in_dir(GConfClient* client, const gchar* dir)
   GSList* pairs;
   GError* error = NULL;
 
-  trace ("Caching values in '%s'\n", dir);
+  trace ("REMOTE: Caching values in '%s'", dir);
   
   PUSH_USE_ENGINE (client);
   pairs = gconf_engine_all_entries(client->engine, dir, &error);
@@ -916,6 +936,8 @@ cache_pairs_in_dir(GConfClient* client, const gchar* dir)
     }
 
   cache_entry_list_destructively (client, pairs);
+  trace ("Mark '%s' as fully cached", dir);
+  g_hash_table_insert (client->cache_dirs, g_strdup (dir), GINT_TO_POINTER (1));
 }
 
 void
@@ -932,7 +954,7 @@ gconf_client_preload    (GConfClient* client,
 #ifdef GCONF_ENABLE_DEBUG
   if (!key_being_monitored (client, dirname))
     {
-      g_warning("Can only preload directories you've added with gconf_client_add_dir() (tried to preload %s)",
+      g_warning("Can only preload directories you've added with gconf_client_add_dir() (tried to preload '%s')",
                 dirname);
       return;
     }
@@ -946,7 +968,7 @@ gconf_client_preload    (GConfClient* client,
 
     case GCONF_CLIENT_PRELOAD_ONELEVEL:
       {
-        trace ("Onelevel preload of '%s'\n", dirname);
+        trace ("Onelevel preload of '%s'", dirname);
         
         cache_pairs_in_dir (client, dirname);
       }
@@ -956,8 +978,9 @@ gconf_client_preload    (GConfClient* client,
       {
         GSList* subdirs;
 
-        trace ("Recursive preload of '%s'\n", dirname);
+        trace ("Recursive preload of '%s'", dirname);
         
+	trace ("REMOTE: All dirs at '%s'", dirname);
         PUSH_USE_ENGINE (client);
         subdirs = gconf_engine_all_dirs(client->engine, dirname, NULL);
         POP_USE_ENGINE (client);
@@ -986,7 +1009,7 @@ gconf_client_set             (GConfClient* client,
 {
   GError* error = NULL;
 
-  trace ("Setting value of '%s'\n", key);
+  trace ("REMOTE: Setting value of '%s'", key);
   PUSH_USE_ENGINE (client);
   gconf_engine_set (client->engine, key, val, &error);
   POP_USE_ENGINE (client);
@@ -1000,8 +1023,7 @@ gconf_client_unset          (GConfClient* client,
 {
   GError* error = NULL;
 
-  trace ("Unsetting '%s'\n", key);
-  
+  trace ("REMOTE: Unsetting '%s'", key);
   PUSH_USE_ENGINE (client);
   gconf_engine_unset(client->engine, key, &error);
   POP_USE_ENGINE (client);
@@ -1022,7 +1044,7 @@ gconf_client_recursive_unset (GConfClient *client,
 {
   GError* error = NULL;
 
-  trace ("Unsetting '%s'\n", key);
+  trace ("REMOTE: Recursive unsetting '%s'", key);
   
   PUSH_USE_ENGINE (client);
   gconf_engine_recursive_unset(client->engine, key, flags, &error);
@@ -1063,16 +1085,31 @@ gconf_client_all_entries    (GConfClient* client,
 {
   GError *error = NULL;
   GSList *retval;
-  
-  trace ("Getting all values in '%s'\n", dir);
+  int dirlen;
+
+  if (g_hash_table_lookup (client->cache_dirs, dir))
+    {
+      GHashTableIter iter;
+      gchar *key;
+      GConfEntry *entry;
+
+      trace ("CACHED: Getting all values in '%s'", dir);
+
+      dirlen = strlen (dir);
+      retval = NULL;
+      g_hash_table_iter_init (&iter, client->cache_hash);
+      while (g_hash_table_iter_next (&iter, &key, &entry))
+        {
+          if (g_str_has_prefix (key, dir) &&
+              key + dirlen == strrchr (key, '/'))
+            retval = g_slist_prepend (retval, gconf_entry_copy (entry));
+        }
+
+      return retval;
+    }
+
+  trace ("REMOTE: Getting all values in '%s'", dir);
 
-  /* We could just use the cache to get all the entries,
-   * iff we have previously done an all_entries and the
-   * cache hasn't since been tossed out, and if we are monitoring
-   * this directory.
-   * FIXME
-   */
-  
   PUSH_USE_ENGINE (client);
   retval = gconf_engine_all_entries (client->engine, dir, &error);
   POP_USE_ENGINE (client);
@@ -1083,8 +1120,12 @@ gconf_client_all_entries    (GConfClient* client,
     return NULL;
 
   if (key_being_monitored (client, dir))
-    cache_entry_list_destructively (client, copy_entry_list (retval));
-  
+    {
+      cache_entry_list_destructively (client, copy_entry_list (retval));
+      trace ("Mark '%s' as fully cached", dir);
+      g_hash_table_insert (client->cache_dirs, g_strdup (dir), GINT_TO_POINTER (1));
+    }
+
   return retval;
 }
 
@@ -1095,7 +1136,7 @@ gconf_client_all_dirs       (GConfClient* client,
   GError* error = NULL;
   GSList* retval;
 
-  trace ("Getting all dirs in '%s'\n", dir);
+  trace ("REMOTE: Getting all dirs in '%s'", dir);
   
   PUSH_USE_ENGINE (client);
   retval = gconf_engine_all_dirs(client->engine, dir, &error);
@@ -1112,7 +1153,7 @@ gconf_client_suggest_sync   (GConfClient* client,
 {
   GError* error = NULL;
 
-  trace ("Suggesting sync\n");
+  trace ("REMOTE: Suggesting sync");
   
   PUSH_USE_ENGINE (client);
   gconf_engine_suggest_sync(client->engine, &error);
@@ -1128,7 +1169,7 @@ gconf_client_dir_exists(GConfClient* client,
   GError* error = NULL;
   gboolean retval;
 
-  trace ("Checking whether directory '%s' exists...\n", dir);
+  trace ("REMOTE: Checking whether directory '%s' exists...", dir);
   
   PUSH_USE_ENGINE (client);
   retval = gconf_engine_dir_exists (client->engine, dir, &error);
@@ -1137,9 +1178,9 @@ gconf_client_dir_exists(GConfClient* client,
   handle_error (client, error, err);
 
   if (retval)
-    trace ("%s exists\n", dir);
+    trace ("'%s' exists", dir);
   else
-    trace ("%s doesn't exist\n", dir);
+    trace ("'%s' doesn't exist", dir);
   
   return retval;
 }
@@ -1156,15 +1197,16 @@ gconf_client_key_is_writable (GConfClient* client,
   g_return_val_if_fail (key != NULL, FALSE);
   g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
 
-  trace ("Checking whether key '%s' is writable... \n", key);
-
   if (gconf_client_lookup (client, key, &entry))
     {
       g_assert (entry != NULL);
 
+      trace ("CACHED: Checking whether key '%s' is writable", key);
       return gconf_entry_get_is_writable (entry);
     }
   
+  trace ("REMOTE: Checking whether key '%s' is writable", key);
+
   entry = get (client, key, TRUE, &error);
 
   if (entry == NULL && error != NULL)
@@ -1179,11 +1221,6 @@ gconf_client_key_is_writable (GConfClient* client,
 
   if (entry)
     gconf_entry_free (entry);
-
-  if (is_writable)
-    trace ("%s is writable\n", key);
-  else
-    trace ("%s is not writable\n", key);
   
   return is_writable;
 }
@@ -1221,9 +1258,10 @@ get (GConfClient *client,
   if (gconf_client_lookup (client, key, &entry))
 
     {
-      trace ("%s was in the client-side cache\n", key);
+      trace ("CACHED: Query for '%s'", key);
       
-      g_assert (entry != NULL);
+      if (entry == NULL)
+        return NULL;
       
       if (gconf_entry_get_is_default (entry) && !use_default)
         return NULL;
@@ -1234,7 +1272,7 @@ get (GConfClient *client,
   g_assert (entry == NULL); /* if it was in the cache we should have returned */
 
   /* Check the GConfEngine */
-  trace ("Doing remote query for %s\n", key);
+  trace ("REMOTE: Query for '%s'", key);
   PUSH_USE_ENGINE (client);
   entry = gconf_engine_get_entry (client->engine, key,
                                   gconf_current_locale(),
@@ -1362,8 +1400,6 @@ gconf_client_get_default_from_schema (GConfClient* client,
   g_return_val_if_fail (GCONF_IS_CLIENT(client), NULL);
   g_return_val_if_fail (key != NULL, NULL);
   
-  trace ("Getting default for %s from schema\n", key);
-  
   /* Check our client-side cache to see if the default is the same as
    * the regular value (FIXME put a default_value field in the
    * cache and store both, lose the is_default flag)
@@ -1371,10 +1407,11 @@ gconf_client_get_default_from_schema (GConfClient* client,
   if (gconf_client_lookup (client, key, &entry))
     {
       g_assert (entry != NULL);
-      
+
       if (gconf_entry_get_is_default (entry))
         {
-          trace ("Using cached value for schema default\n");
+	  trace ("CACHED: Getting schema default for '%s'", key);
+
           return gconf_entry_get_value (entry) ?
             gconf_value_copy (gconf_entry_get_value (entry)) :
             NULL;
@@ -1382,7 +1419,7 @@ gconf_client_get_default_from_schema (GConfClient* client,
     }
 
   /* Check the GConfEngine */
-  trace ("Asking engine for schema default\n");
+  trace ("REMOTE: Getting schema default for '%s'", key);
   PUSH_USE_ENGINE (client);
   val = gconf_engine_get_default_from_schema (client->engine, key,
                                               &error);
@@ -1682,7 +1719,7 @@ gconf_client_set_float   (GConfClient* client, const gchar* key,
   g_return_val_if_fail(GCONF_IS_CLIENT(client), FALSE);  
   g_return_val_if_fail(key != NULL, FALSE);
 
-  trace ("Setting float %s\n", key);
+  trace ("REMOTE: Setting float '%s'", key);
   PUSH_USE_ENGINE (client);
   result = gconf_engine_set_float (client->engine, key, val, &error);
   POP_USE_ENGINE (client);
@@ -1707,7 +1744,7 @@ gconf_client_set_int     (GConfClient* client, const gchar* key,
   g_return_val_if_fail(GCONF_IS_CLIENT(client), FALSE);  
   g_return_val_if_fail(key != NULL, FALSE);
 
-  trace ("Setting int %s\n", key);
+  trace ("REMOTE: Setting int '%s'", key);
   PUSH_USE_ENGINE (client);
   result = gconf_engine_set_int (client->engine, key, val, &error);
   POP_USE_ENGINE (client);
@@ -1733,7 +1770,7 @@ gconf_client_set_string  (GConfClient* client, const gchar* key,
   g_return_val_if_fail(key != NULL, FALSE);
   g_return_val_if_fail(val != NULL, FALSE);
 
-  trace ("Setting string %s\n", key);
+  trace ("REMOTE: Setting string '%s'", key);
   PUSH_USE_ENGINE (client);
   result = gconf_engine_set_string(client->engine, key, val, &error);
   POP_USE_ENGINE (client);
@@ -1758,6 +1795,7 @@ gconf_client_set_bool    (GConfClient* client, const gchar* key,
   g_return_val_if_fail(GCONF_IS_CLIENT(client), FALSE);  
   g_return_val_if_fail(key != NULL, FALSE);
 
+  trace ("REMOTE: Setting bool '%s'", key);
   PUSH_USE_ENGINE (client);
   result = gconf_engine_set_bool (client->engine, key, val, &error);
   POP_USE_ENGINE (client);
@@ -1783,7 +1821,7 @@ gconf_client_set_schema  (GConfClient* client, const gchar* key,
   g_return_val_if_fail(key != NULL, FALSE);
   g_return_val_if_fail(val != NULL, FALSE);
 
-  trace ("Setting schema %s\n", key);
+  trace ("REMOTE: Setting schema '%s'", key);
   PUSH_USE_ENGINE (client);
   result = gconf_engine_set_schema(client->engine, key, val, &error);
   POP_USE_ENGINE (client);
@@ -1810,7 +1848,7 @@ gconf_client_set_list    (GConfClient* client, const gchar* key,
   g_return_val_if_fail(GCONF_IS_CLIENT(client), FALSE);  
   g_return_val_if_fail(key != NULL, FALSE);
 
-  trace ("Setting list %s\n", key);
+  trace ("REMOTE: Setting list '%s'", key);
   PUSH_USE_ENGINE (client);
   result = gconf_engine_set_list(client->engine, key, list_type, list, &error);
   POP_USE_ENGINE (client);
@@ -1838,7 +1876,7 @@ gconf_client_set_pair    (GConfClient* client, const gchar* key,
   g_return_val_if_fail(GCONF_IS_CLIENT(client), FALSE);  
   g_return_val_if_fail(key != NULL, FALSE);
 
-  trace ("Setting pair %s\n", key);
+  trace ("REMOTE: Setting pair '%s'", key);
   PUSH_USE_ENGINE (client);
   result = gconf_engine_set_pair (client->engine, key, car_type, cdr_type,
                                   address_of_car, address_of_cdr, &error);
@@ -1915,7 +1953,7 @@ gconf_client_cache (GConfClient *client,
 
       if (changed)
         {
-          trace ("Updating value of '%s' in the cache\n",
+          trace ("Updating value of '%s' in the cache",
                  new_entry->key);
 
           if (!take_ownership)
@@ -1934,7 +1972,7 @@ gconf_client_cache (GConfClient *client,
         }
       else
         {
-          trace ("Value of '%s' hasn't actually changed, would have updated in cache if it had\n",
+          trace ("Value of '%s' hasn't actually changed, would have updated in cache if it had",
                  new_entry->key);
 
           if (take_ownership)
@@ -1950,7 +1988,7 @@ gconf_client_cache (GConfClient *client,
         new_entry = gconf_entry_copy (new_entry);
       
       g_hash_table_insert (client->cache_hash, new_entry->key, new_entry);
-      trace ("Added value of '%s' to the cache\n",
+      trace ("Added value of '%s' to the cache",
              new_entry->key);
 
       return TRUE; /* changed */
@@ -1970,7 +2008,26 @@ gconf_client_lookup (GConfClient *client,
   entry = g_hash_table_lookup (client->cache_hash, key);
 
   *entryp = entry;
-      
+
+  if (!entry)
+  {
+    char *dir, *last_slash;
+
+    dir = g_strdup (key);
+    last_slash = strrchr (dir, '/');
+    g_assert (last_slash != NULL);
+    *last_slash = 0;
+
+    if (g_hash_table_lookup (client->cache_dirs, dir))
+    {
+      g_free (dir);
+      trace ("Negative cache hit on %s", key);
+      return TRUE;
+    }
+
+    g_free (dir);
+  }
+
   return entry != NULL;
 }
 
@@ -2350,7 +2407,7 @@ static void
 gconf_client_queue_notify (GConfClient *client,
                            const char  *key)
 {
-  trace ("Queing notify on %s, %d pending already\n", key,
+  trace ("Queing notify on '%s', %d pending already", key,
          client->pending_notify_count);
   
   if (client->notify_handler == 0)
@@ -2423,7 +2480,7 @@ gconf_client_flush_notifies (GConfClient *client)
   GSList *to_notify;
   GConfEntry *last_entry;
 
-  trace ("Flushing notify queue\n");
+  trace ("Flushing notify queue");
   
   /* Adopt notify list and clear it, to avoid reentrancy concerns.
    * Sort it to compress duplicates, and keep people from relying on
@@ -2445,18 +2502,18 @@ gconf_client_flush_notifies (GConfClient *client)
         {
           if (entry != last_entry)
             {
-              trace ("Doing notification for %s\n", entry->key);
+              trace ("Doing notification for '%s'", entry->key);
               notify_one_entry (client, entry);
               last_entry = entry;
             }
           else
             {
-              trace ("Ignoring duplicate notify for %s\n", entry->key);
+              trace ("Ignoring duplicate notify for '%s'", entry->key);
             }
         }
       else
         {
-          trace ("Key %s was in notify queue but not in cache; we must have stopped monitoring it; not notifying\n",
+          trace ("Key '%s' was in notify queue but not in cache; we must have stopped monitoring it; not notifying",
                  tmp->data);
         }
       
